;; -*- mode: common-lisp; package: user -*-

;;; 
;;; File: jqp-lib.cl
;;;
;;; Author: Sung-youn Kim (syounkim@gmail.com)
;;;
;;; Copyright: All rights are reserved
;;;  
;;; This file contains basic functions and utilities for output, 
;;; running simulation, and building knowledge structure from
;;; data of a certain format (alist), and so on.
;;;
;;;


(in-package #:user)


;;;
;;;
;;; Global Variables
;;;
;;;


;; for output

(defvar *output-dir* nil) 

(defvar *agent-internal-output-file* nil)

(defvar *agent-survey-output-evaluation-file* nil)

(defvar *agent-survey-output-perception-file* nil) 

;; for simulation

(defvar *simulation-spec-agent-id* nil)

(defvar *jqp-agent-id* nil)

(defvar *jqp-group-id* nil)

(defvar *update-spec* nil)

(defvar *update-value* nil)


;;;
;;;
;;; Collect and Print Result into Files
;;;
;;;


(defun jqp-output ()
  "Given *jqp-act-out*, which holds all transient activation
levels and positive and negative affects of all chunks with time, it
returns them for chunks in *chunks-to-check* in *jqp-out*."
  (setf *jqp-out* nil)
  (dotimes (i (length *jqp-act-out*) t)
    (let* ((x (nth i *jqp-act-out*))
	   (time (caar (last x)))
	   (results nil))     
      (push (list time) results)
      (dolist (wme *chunks-to-check* t)
	(setf wme (get-safe-wme wme))
	(when (and (nth 1 (assoc wme x))                
		   (null (nth 2 (assoc wme x)))
		   (null (nth 3 (assoc wme x)))
		   (null (nth 4 (assoc wme x))))
	    (dotimes (j 3 t)
	      (setf (nth (+ j 2) (assoc wme x)) (nth 1 (assoc wme x)))))
	;; this part (when ...) is for inserting b.act for s, sac, and
	;; tact, which are nil for the records of activation generated
	;; after each info is processed.
	(push (assoc wme x) results))
      (push (reverse results) *jqp-out*))))

(unless (compiled-function-p #'jqp-output) (compile 'jqp-output)) 


(defun format-float (num)
  "It formats float number representation such that d0, d-4, etc are removed."
  (format nil "~10,3f" num))


(defun jqp-print-survey-response (&key mysql-db)
  "It prints out survey responses."
  (let ((if-exist (if mysql-db :supersede :append)))
    (with-open-file (stream *agent-survey-output-evaluation-file*
		     :direction :output :if-exists if-exist :if-does-not-exist :create) 
      (let ((survey (reverse *survey-response-evaluation*)))
	(do ((j 0 (+ j 1)))
	    ((> j (- (length survey) 1)) nil)
	  (let ((time (nth 0 (nth j survey)))
		(obj (nth 1 (nth j survey)))
		(res (nth 2 (nth j survey))))
	    (format stream "~a,~a,~a,~a,~a,~a,~%" *simulation-id* *jqp-agent-id* 
		    *jqp-group-id* time obj (format-float res))))))))


(defun jqp-print-survey-response-perception (&key mysql-db)
  "It prints out survey perception responses. Experimental yet."
  (let ((if-exist (if mysql-db :supersede :append)))
    (with-open-file (stream *agent-survey-output-perception-file*
		     :direction :output :if-exists if-exist :if-does-not-exist :create)
      (let ((survey (reverse *survey-response-perception*)))
	(do ((j 0 (+ j 1)))
	    ((> j (- (length survey) 1)) nil)
	  (let ((time (nth 0 (nth j survey)))
		(obj (nth 1 (nth j survey)))
		(trait (nth 2 (nth j survey)))
		(res1 (nth 3 (nth j survey)))
		(res2 (nth 4 (nth j survey)))
		(res3 (nth 5 (nth j survey))))
	    (format stream "~a,~a,~a,~a,~a,~a,~a,~a,~a,~%" 
		    *simulation-id* *jqp-agent-id* *jqp-group-id* time obj trait 
		    (format-float res1) (format-float res2) (format-float res3)
		    )))))))


(defun jqp-voting-print-output (&key mysql-db)
  "It prints out all activation levels and affect over time for chunks
specified in *chunks-to-check* into *agent-internal-output-file*."
 (let ((if-exist (if mysql-db :supersede :append)))
   (with-open-file (stream *agent-internal-output-file*
		    :direction :output :if-exists if-exist :if-does-not-exist :create) 
     (dotimes (j (length *jqp-out*))
       (let* ((time-wmes (nth j *jqp-out*))
	      (time (car (nth 0 time-wmes)))
	      (wmes (cdr time-wmes)))
	 (dolist (wme *chunks-to-check*)
	   (setf wme (get-safe-wme wme))
	   (when (assoc wme wmes)
	     (let ((wme-name (car (assoc wme wmes)))
		   (data (cdr (assoc wme wmes))))
	       (format stream "~A,~A,~A,~A,~A,~A,~A,~A,~A,~A,~A,~A,,~%"
		       *simulation-id* *jqp-agent-id* *jqp-group-id*
		       time wme-name 
		       (format-float (nth 0 data))
		       (format-float (nth 1 data))
		       (format-float (nth 2 data))
		       (format-float (nth 3 data))
		       (format-float (nth 4 data))
		       (format-float (nth 5 data))
		       (format-float (+ (nth 4 data) (nth 5 data))))))))))))


;;;
;;;
;;; Simulation 
;;;
;;;


(defun gen-alist-total-num-per-group (&key (total-num 100) num-per-group)
  "It generates an alist that specifies how many agents per group will be 
employed in a simulation. The proportion is based on 2000 ANES survey data."
  (list (list 'strong-conservative (if num-per-group num-per-group (round (* total-num 0.0681))))
	(list 'conservative (if num-per-group num-per-group (round (* total-num 0.2941))))
	(list 'moderate (if num-per-group num-per-group (round (* total-num 0.4051))))
	(list 'liberal (if num-per-group num-per-group (round (* total-num 0.1913))))
	(list 'strong-liberal (if num-per-group num-per-group (round (* total-num 0.0419))))))


(defun gen-alist-agent-ind-group-id (group-num var-name) 
  "It generates an alist that specifies individual id and group id for each and
  every agent to be employed in a sumulation. It is used to determine 
*simulation-spec-agent-id* for group level simulation."    
  (let ((count 0) out) (dolist (i group-num) (let
  ((group (first i)) (num (second i))) (dotimes (n num) (incf count)
  (push (list count group) out)))) (list (reverse out) var-name)))


(defun get-agent-ind-group-id (data subjid-order group)
  "It retrieves an agent's individual and group id from 
*simulation-spec-agent-id*."
  (let ((subjid-list (mapcar #'first (car data))))
    (list (nth (- subjid-order 1) subjid-list) 
	  (jqp-get-data-from-table (nth (- subjid-order 1) subjid-list) group data))))


(defun jqp-agent-initialize (id group model-file) 
  "It sets up relevant variables and parameters. It takes command line arguments 
for agent id."  
  (progn			
    (setf *output-dir* (namestring (current-directory))) 
    (let ((a (get-agent-ind-group-id *simulation-spec-agent-id* id group))) 
      (setf *jqp-agent-id* (first a)) 
      (setf *jqp-group-id* (second a)))
    (progn (setf *agent-internal-output-file* (namestring (merge-pathnames *output-dir* (format nil "voting-internal-output-agent-~a-~a" *jqp-group-id* *jqp-agent-id*))))
	   (when (probe-file *agent-internal-output-file*) (delete-file *agent-internal-output-file*)))
    (progn (setf *agent-survey-output-evaluation-file* (namestring (merge-pathnames *output-dir* (format nil "voting-survey-output-agent-~a-~a" *jqp-group-id* *jqp-agent-id*))))
	   (when (probe-file *agent-survey-output-evaluation-file*) (delete-file *agent-survey-output-evaluation-file*)))
    (progn (setf *agent-survey-output-perception-file* (namestring (merge-pathnames *output-dir* (format nil "voting-survey-output-perception-agent-~a-~a" *jqp-group-id* *jqp-agent-id*))))
	   (when (probe-file *agent-survey-output-perception-file*) (delete-file *agent-survey-output-perception-file*)))
    (load model-file)))


(defun do-simulation (init-parameter update-spec update-increment 
			     &key final-run (time 3) (sim-start 1) (in-order t))
  "It runs simulation over different parameter values. Initial parameter values 
and their changes are determined by update-spec and update-increment."
  (setf *simulation-id* sim-start)
  (setf *update-spec* update-spec)
  (setf *update-value* update-increment)
  (jqp-set-parameters :rho (nth 0 init-parameter) :gamma (nth 1 init-parameter) 
		      :delta (nth 2 init-parameter) :sigma (nth 3 init-parameter))
  (let ((init-anchor *rho*)
	(init-affect-congruence *gamma*)
	(init-color-degree *delta*)
	(init-noise *sigma*))
    (do ((i 1 (+ i 1)))
	((> i (nth 0 *update-spec*)) (setf *rho* init-anchor))    
      (do ((j 1 (+ j 1)))
	  ((> j (nth 1 *update-spec*)) (setf *gamma* init-affect-congruence))      
	(do ((k 1 (+ k 1)))
	    ((> k (nth 2 *update-spec*)) (setf *delta* init-color-degree))
	  (do ((u 1 (+ u 1)))
	      ((> u (nth 3 *update-spec*)) (setf *sigma* init-noise))
	    (progn (jqp-reset)
		   (do-experiment time :final-run final-run :in-order in-order)
		   (incf *simulation-id*))
	    (progn (incf *sigma* (nth 3 *update-value*))
		   (progn (sgp-fct (list :ans *sigma*))
			  (sgp-fct (list :pas *sigma*)))))	  
	  (incf *delta* (nth 2 *update-value*)))
	(incf *gamma* (nth 1 *update-value*)))
      (incf *rho* (nth 0 *update-value*)))))


;;;
;;;
;;; Building initial knowledge from data in alists
;;;
;;;


(defun jqp-get-data-from-table (rowname colname table)
  (let ((data (assoc rowname (first table) :test #'equalp))
	(var-name (second table)))
    (if (and data (member colname var-name))
	(nth (position colname var-name :test #'equalp) data))))


(defun jqp-build-initial-knowledge-with-noise-from-alist 
    (wmes-alist &key (sd :empirical) (type 'meaning) group issue (check-duplicate nil))
  "Given alist of initial knowledges that contains wme names, mean postive 
and negative affect, and variances, it puts them into LTM such that the affect 
is normally distributed with the mean and variance."
  (let ((wmes-alist (if group (rest (assoc group wmes-alist)) wmes-alist)))
    (dolist (wme-alist wmes-alist)
      (let* ((wme-name (nth 0 wme-alist))
	     (pmean (nth 1 wme-alist))
	     (nmean (nth 2 wme-alist))
     	     (psd (if (eq sd :empirical) (jqp-convert-sd (nth 3 wme-alist)) sd))
	     (nsd (if (eq sd :empirical) (jqp-convert-sd (nth 4 wme-alist)) sd))
	     (pos (+ pmean (noise psd)))
	     (neg (+ nmean (noise nsd))))	
	(if check-duplicate
	    (when (null (get-safe-wme wme-name))
	      (jqp-add-dm wme-name pos neg :type type :issue issue))
	  (jqp-add-dm wme-name pos neg :type type :issue issue))))))


(defun jqp-build-initial-knowledge-association-from-alist 
    (assoc-alist &key group issue (scalor 1) threshold)
  "Given an alist that contains the strenghts of associations between chunks, 
it sets association b/w chunks in memory."
  (let ((assoclist (if group (rest (assoc group assoc-alist)) 
		     assoc-alist)))
    (dolist (i assoclist)
      (let* ((wmei (get-safe-wme (nth 0 i)))
	     (wmejs (rest i)))
	(dolist (j wmejs)
	  (let ((wmej (get-safe-wme (nth 0 j)))
		(num  (* (nth 1 j) (nth 3 j))))
	    (when (and threshold (>= num threshold)) 
	      (jqp-add-connection wmei wmej (round (* scalor num))) 
	      (when issue 
		(let ((anti-wmej 
		       (get-safe-wme (intern (concatenate 'string "ANTI-" (symbol-name (nth 0 j))))))
		      (num1 (* (- 1 (nth 1 j)) (nth 3 j))))
		  (jqp-add-connection wmei anti-wmej (round (* scalor num1))) 
		  )))))))))


(defun jqp-set-initial-ol-order-from-alist 
    (wmes-alist order &key set-order group issue (scalor 1))  
  "Given an alist that contains paffect, naffect, psd, nsd and response rate, 
it sets their initial ol-order by setting ia-account between each chunk and 
a system chunk, init-ol-chunk, based on response rate."
  (let ((ol-list (if group (rest (assoc group wmes-alist))
		     wmes-alist)))
      (dolist (i ol-list)
	(let ((wmej (get-safe-wme 'init-ol-order))
              (wmei (get-safe-wme (nth 0 i)))
	      (ol-order (if set-order set-order 
			  (let ((scaled-ol-order (* order scalor)))
			    (round (* scaled-ol-order (nth 5 i)))))))
	  (progn (setf (ia-count (get-make-ia wmej wmei)) ol-order)
		 (eval (list 'set-ia 
			     (list wmej wmei 0))))		 
	  (when issue 
	    (let ((anti-wmei 
		   (get-safe-wme 
		    (intern (concatenate 'string "ANTI-" (symbol-name (nth 0 i)))))))
	      (progn (setf (ia-count (get-make-ia wmej anti-wmei)) ol-order)
		     (eval (list 'set-ia 
				 (list wmej anti-wmei 0))))))))))


(defun jqp-set-base-level-from-alist 
    (wmes-alist &optional (creation-time nil) &key group base-reference issue scalor)
  "Given an alist that contains paffect, naffect, psd, nsd and response rate, 
it sets their baseline activation based on response rate."
  (let ((wmes-list (if group (rest (assoc group wmes-alist))
		     wmes-alist)))
      (dolist (i wmes-list)
	(let ((wme (get-safe-wme (nth 0 i)))
	      (base-level (if scalor 
			      (if (numberp (nth 5 i))
				  (- base-reference (round (* scalor (- 1 (nth 5 i)))))
				(error (format t "Response rate is not a number")))
			    base-reference)))
	  (set-base-level wme (list base-level creation-time))		 
	  (when issue 
	    (let ((anti-wme 
		   (get-safe-wme 
		    (intern (concatenate 'string "ANTI-" 
					 (symbol-name (nth 0 i)))))))
	      (set-base-level anti-wme (list base-level creation-time))))))))


;;;
;;;
;;; Others and Utilities
;;;
;;;


(defun jqp-sum (x)
  "It returns the sum of a list of numbers as a number"
  (cond ((null x) 0)
	((numberp (first x)) (+ (first x) (jqp-sum (rest x))))
	(t (jqp-sum (rest x)))))

(defun jqp-max (x) 
  "It returns the maximum number in a list of numbers" 
  (cond ((null x) most-negative-double-float)
	((numberp (first x))
	 (max (first x) (jqp-max (rest x))))
	((listp (first x))
	 (error (format t "~s is not a number" (first x))))
	(t (jqp-max (rest x)))))

(defun jqp-min (x) 
  "It returns the maximum number in a list of numbers" 
  (cond ((null x) most-positive-double-float)
	((numberp (first x))
	 (min (first x) (jqp-min (rest x))))
	((listp (first x))
	 (error (format t "~s is not a number" (first x))))
	(t (jqp-min (rest x)))))


(defun jqp-product (x) 
   "It returns the product of a list of numbers as a number" 
   (cond ((null x) 1)
	 ((numberp (first x))
	  (* (first x) (jqp-product (rest x))))
	 ((listp (first x))
	  (error (format t "~s is not a unmber" x)))
	 (t (jqp-product (rest x)))))


(defun jqp-average (x)
  "It returns the average of a list of numbers as a number"
  (* 1.0 (/ (jqp-sum x) (length x))))


(defun jqp-logit (x &key (m 0) (s 1))
  (/ 1 (+ 1 (exp (/ (* (- x m) -1) s)))))

(defun jqp-truncate-negative (x)
  (mapcar #'(lambda (n) 
	      (if (>= n 0) n 0)) x))

(defun jqp-sublist (list x y)
  (let ((lis nil))
    (do ((i (- x 1) (+ i 1)))
	((eq i y) (reverse lis))
      (push (nth i list) lis))))


(defun jqp-convert-sd (sd)
  "It rescales standard deviation into corresponding 
s-value (/ (sqrt (* 3.0 variance)) 3.1416) for the noise 
function in ACT-R."
  (/ (sqrt (* 3.0 (expt sd 2))) 3.1416))


(defun jqp-dump-image (dxl-file)
  (progn (excl:gc t)
	 (excl:dumplisp :name (string dxl-file)))) 


(defun suppress-output ()
  (setf *standard-output* 
    (open "/dev/null" :direction :output :if-exists :supersede))
  (setf *error-output* 
    (open "/dev/null" :direction :output :if-exists :supersede)))


(defun study-sentence-model (text time)
  (let ((window (open-exp-window "Processing Campaign Information" :visible nil :width 500)))
    (pm-install-device window)
    (add-text-to-exp-window :text text :x 25 :y 150 :width 250)
    (pm-proc-display :clear t)
    (pm-run time :full-time t)))

(defun study-sentence-person (text time)
  (let ((window (open-exp-window "Campaign Information" :visible nil :width 500)))
    (add-text-to-exp-window :text text :x 25 :y 150 :width 250)
    (sleep time)))
