;; -*- mode: common-lisp; package: user -*-

;;; 
;;; File: jqp.cl
;;;
;;; Author: Sung-youn Kim (syounkim@gmail.com)
;;;
;;; Copyright: All rights are reserved
;;;
;;; This file provides fundamental mechanisms for JQP. It includes
;;; mechanisms for knowledge and attitude representation, affective
;;; priming, memory-based processing, and on-line processing
;;;
;;;


(in-package #:user)


;;;
;;; Global Variables and Parameters 
;;;


(defvar *rho* nil
"It governs the degree of influence from new vs old information.") 

(defvar *gamma* nil
"It governs the degree of affective priming effect.")

(defvar *delta* nil
  "The degree to which considerations on the spot influence  
explicit evaluation of object.")

(defvar *sigma* nil 
  "Random noise,") 

(defvar *jqp-act-hold* nil
  "Stores baselevel, spreading activation, and affective 
priming effect in connection with test-slots-pm")

(defvar *jqp-transient-activation-holding* nil
  "Temporarily holds all transient actvation levels, that is, 
baselevel, spreading activation, affective priming in 
connection with jqp-get-transient-activation and 
jqp-get-activation-fct")

(defvar *affect-holding* nil
  "It temporarily holds affect of wmes in WM, It is used in 
affective-congruence")

(defvar *similarity-string* 5
  "The additional activation that wmes get when they match 
external information (string) being seen by visual module.")

(defvar *similarity-string-penalty* -5
  "The decrease in activation that wmes lose when they do not 
match external information (string) being seen by visual module.")

(defvar *chunks-to-check* nil
  "The list of chunks to be selected from *jqp-act-out* to be 
included in *jqp-out*. It should be set in model file.")

(defvar *jqp-act-out* nil
  "It stores all changes in activation levels over time
 with time stamps")

(defvar *jqp-out* nil
  "It stores everything, all activation levels and positive 
and negative affect for each time, for chunks in *chunks-to-check* 
via jqp-output")

(defvar *survey-response-evaluation* nil
  "It stores the constructed evaluation of objects.")

(defvar *survey-response-perception* nil
"The perceptions of political objects' traits and issue positions.")


;;;
;;; Basic Functions
;;;


(defun jqp-set-slot-value (wme slot newvalue)
  "It sets the value of slot in wme to newvalue and compute ia again 
depending on :compute-ia."
  (let ((index (get-safe-slot slot (wme-type wme) wme)))
    (set-slot-value wme (slot-index index) newvalue :compute-ia t)))


(defun jqp-get-ia (wmej wmei)
  "It prints out the ia and ia counts between wmej and wmei."
  (let* ((wmei (get-safe-wme wmei))
	 (wmej (get-safe-wme wmej))
	 (ia (get-ia wmej wmei))
	 (ia-c (ia-count ia)))
    (list ia ia-c)))

(defun jqp-add-connection (wmej wmei num)
  "add num connections from wmei to wmej, which strengthens Sji 
while weakens Sjk due to the fan effect. It also sets the 
Sij = 0, only when it is null."
  (let ((wmej (get-safe-wme wmej))
	(wmei (get-safe-wme wmei)))
    (dotimes (i num) 
      (add-connection wmej wmei))
    (when (null (get-ia (get-safe-wme wmei) (get-safe-wme wmej)))
      (eval (list 'set-ia 
		  (list wmei wmej 0))))))


(defun jqp-get-safe-anti-wme (name)
  "Gets the wme corresponding to anti-name.  Prints an error message if undefined."
  (let ((name (read-from-string (format nil "anti-~s" name))))
    (get-safe-wme name)))


(defun jqp-set-parameters (&key rho gamma delta sigma)
  (progn (setf *rho* rho)    
	 (setf *gamma* gamma)       
	 (setf *delta* delta) 
	 (progn (setf *sigma* sigma) 
		(sgp-fct (list :ans *sigma*))
		(sgp-fct (list :pas *sigma*)))))



;;;
;;;
;;; Representation of Affect in LTM (Hot Cognition)
;;;
;;;


(defun jqp-make-chunk-type () 
  (chunk-type meaning positive negative purpose word state)
  (chunk-type comprehend-sentence positive negative positive1 negative1 
	      agent action object source warning purpose word state) 
  (chunk-type report-attitude positive negative agent action object 
	      purpose word state)
  (chunk-type report-perception positive negative agent action object 
	      purpose word state))


(defun jqp-add-goal ()
  (add-dm 
    (goal ISA comprehend-sentence positive 0 negative 0 positive1 0 negative1 0 purpose "study-sentence") 
    (goal-RA ISA report-attitude positive 0 negative 0 purpose "report-attitude")
    (goal-RP ISA report-perception positive 0 negative 0 purpose "report-perception")))


(defun jqp-add-dm (wme-name pos neg &key (type 'meaning) issue)
  "Given the mean of postive and negative affect for a concept, it puts 
it into memory with the affect."
  (eval (list 'add-dm (list wme-name 'ISA type 'word 
			    (remove #\- (string-downcase (symbol-name wme-name))))))
  (let ((wme (get-safe-wme wme-name)))
    (cond ((and (> pos 0) (< neg 0))
	   (progn (jqp-set-slot-value wme 'positive pos)
		  (jqp-set-slot-value wme 'negative neg)))
	  ((and (> pos 0) (> neg 0))
	   (progn (jqp-set-slot-value wme 'positive (+ pos neg))
		  (jqp-set-slot-value wme 'negative 0)))
	  ((and (< pos 0) (> neg 0))
	   (progn (jqp-set-slot-value wme 'positive neg)
		  (jqp-set-slot-value wme 'negative pos)))
	  ((and (< pos 0) (< neg 0))
	   (progn (jqp-set-slot-value wme 'positive 0)
		  (jqp-set-slot-value wme 'negative (+ pos neg))))
	  (t
	   (progn (jqp-set-slot-value wme 'positive pos)
		  (jqp-set-slot-value wme 'negative neg)))))
  (when issue
    (let ((anti-wme-name (intern (concatenate 'string "ANTI-" (symbol-name wme-name)))))
      (eval (list 'add-dm (list anti-wme-name 'ISA type 'word 
				(remove #\- (string-downcase (symbol-name anti-wme-name))))))
      (let ((wme (get-safe-wme wme-name))
	    (anti-wme (get-safe-wme anti-wme-name)))
	(jqp-set-slot-value anti-wme 'positive (* -1 (get-safe-slot-value wme 'negative)))
	(jqp-set-slot-value anti-wme 'negative (* -1 (get-safe-slot-value wme 'positive)))))))


;;;
;;;
;;; Affective Congruence and Priming
;;;
;;;

(defun jqp-distance (x y)
  "In conjuction with affective-congruence, it determines the 
affective congruence b/w wmei in WM (source) and wmej in LTM."
  (if (= (* (+ (nth 0 x) (nth 1 x)) (+ (nth 0 y) (nth 1 y))) 0) 0 1))


(defun affective-congruence (y x) 
  "Together with other built-in mechanisms in ACT-R, it implements Eq (1).
 This is the hook function used along with *similarity-hook-fn*. 
It determines the activation increase/decrease for a wme in LTM due to 
1) its affective congruence with wme in WM and/or 2) matching external 
information (a string). For matching external info, its degree is 
influenced by *similarity-string* and *similarity-string-penalty*. 
Affective congruence is determined in a binary fashon (1, 0, -1) and 
*gamma* determines the magnitude of affective priming effect."  
  (cond ((and (stringp x) (stringp y))
	 (if (equalp x y) *similarity-string* *similarity-string-penalty*))
	((and (listp x) (numberp y))
	 (let ((affect *affect-holding*) (outcome nil))
	   (if (and (listp affect) affect)	     
	       (do ((i 0 (+ 1 i))
		    (j 0 (+ 1 j)))
		   ((or (> i (- (length x) 1))
			(> j (- (length affect) 1)))
		    (progn (setf *affect-holding* nil) 
			   (jqp-average outcome)))
		 (let* ((positive-source (first (nth j affect)))
			(positive-chunk (second (nth j affect)))
			(negative-source (nth i x))
			(negative-chunk y)
			(attitude-source (+ positive-source negative-source))
			(attitude-chunk (+ positive-chunk negative-chunk))
			(dist (jqp-distance (list positive-source negative-source) 
					    (list positive-chunk negative-chunk)))
			(aff-cong    
			 (cond
			  ((= (* attitude-source attitude-chunk) 0) 0)
			  ((< (* attitude-source attitude-chunk) 0) 
			   (* *gamma* (* -1 dist)))
			  ((> (* attitude-source attitude-chunk) 0) 
			   (* *gamma* dist))
			  (t 0))))
		   (push aff-cong outcome)))
	     (when (null affect)
	       (do ((i 0 (+ 1 i)))
		   ((> i (- (length x) 1)) 
		    (progn (setf *affect-holding* (reverse outcome)) 0))
		 (push (list (nth i x) y) outcome))))))
	((and (numberp x) (numberp y))
	 (let ((affect *affect-holding*))
	   (if affect	     
	       (let* ((positive-source (first affect))
		      (positive-chunk (second affect))
		      (negative-source x)
		      (negative-chunk y)
		      (attitude-source (+ positive-source negative-source))
		      (attitude-chunk (+ positive-chunk negative-chunk))
		      (dist (jqp-distance (list positive-source negative-source) 
					  (list positive-chunk negative-chunk))))
		 (setf *affect-holding* nil)
		 (cond
		  ((= (* attitude-source attitude-chunk) 0) 0)
		  ((< (* attitude-source attitude-chunk) 0) 
		   (* *gamma* (* -1 dist)))
		  ((> (* attitude-source attitude-chunk) 0) 
		   (* *gamma* dist))
		  (t 0)))		 
	     (when (null affect)
	       (setf *affect-holding* (list x y)) 0))))
	(t 0)))	


;;;
;;;    
;;;  Memory-based Processing - Evaluation (Attitude) Construction
;;;
;;;


(defun jqp-get-associated-chunks-list-fct (wme &key slot-value)
  "It returns a list of chunks in LTM associated with wme (including wme itself). 
If slot-value is provided, it returns only those with the slot-value 
among the associated wmes."
  (if (listp wme) wme 
    (let ((ias (wme-ias (get-safe-wme wme)))
	  (slot (when slot-value (first slot-value)))
	  (s-val (when slot-value (second slot-value)))
	  (wmes nil))
      (dolist (i ias wmes) 
	(let* ((x (get-safe-wme (car i)))
	       (x-slot-val (when (and slot s-val) (get-slot-value x slot)))) 
	  (unless (equal x (get-safe-wme 'init-ol-order))
	    (if x-slot-val (when (equal x-slot-val s-val) (push x wmes))
	      (push x wmes))))))))

 
(defun jqp-get-considerations-on-the-spot (wmes &key slot-value)
  "Returns the list of chunks in LTM associated with wmes using 
jqp-get-associated chunks-list-fct"
  (if (listp wmes)
      (let ((results nil))
	(dolist (i wmes (remove-duplicates results))
	  (let ((considerations 
		 (remove (get-safe-wme i) 
			 (jqp-get-associated-chunks-list-fct i :slot-value slot-value))))
	    (when considerations 
	      (setf results (append considerations results))))))
      (error (format t "The argument is not a list."))))


(defun jqp-get-paffect-associated-chunks-fct (wme &key slot-value)
  "Returns the positive affect of chunks associated with wme."
  (let ((results nil)
	(wmes (jqp-get-associated-chunks-list-fct wme :slot-value slot-value)))
    (dolist (i wmes (reverse results))
	(setf i (get-safe-wme i))
	(when i
	  (push (get-safe-slot-value i 'positive) results)))))


(defmacro jqp-get-paffect-associated-chunks (wme)
  `(jqp-get-paffect-associated-chunks-fct ',wme))


(defun jqp-get-naffect-associated-chunks-fct (wme &key slot-value)
  "Returns the negative affect of chunks associated with wme."  
  (let ((results nil)
	(wmes (jqp-get-associated-chunks-list-fct wme :slot-value slot-value)))
    (dolist (i wmes (reverse results))
	(setf i (get-safe-wme i))
	(when i
	  (push (get-safe-slot-value i 'negative) results)))))


(defmacro jqp-get-naffect-associated-chunks (wme)
  `(jqp-get-naffect-associated-chunks-fct ',wme))


;;
;; Getting Transient Activation Level
;; 
;; This part defines jqp-get-transient-activation and puts it into
;; compute-base-level-activation and test-slots-pm defined in
;; actr5.lisp. It gets and records activation levels, which are then
;; used by attitude construction functions. This part should be placed
;; in actr5.lisp
;;


(defun jqp-get-transient-activation (time &key wme b-act s-act sac-act t-act)
  "It stores in *jqp-transient-activation-holding* and
*jqp-act-out* 1) base-level, 2) base-level + spreading
activation, 3) base-level + spreading activation + affective
congruence, 4) base-level + spreading activation + affective
congruence + matching external info, 5) positive affect, 
and 6) negative affect"
  (let ((pos (get-safe-slot-value (get-safe-wme wme) 'positive))
	(neg (get-safe-slot-value (get-safe-wme wme) 'negative)))	
    (if *jqp-transient-activation-holding*
	(if (eql time (caar (last (car *jqp-transient-activation-holding*))))
	    (if (null (assoc wme (car *jqp-transient-activation-holding*)))
		(push (list wme b-act s-act sac-act t-act pos neg) (car *jqp-transient-activation-holding*))
	      (cond ((and (null (nth 1 (assoc wme (car *jqp-transient-activation-holding*)))) b-act)		  
		     (setf (nth 1 (assoc wme (car *jqp-transient-activation-holding*))) b-act))
		    ((and (null (nth 2 (assoc wme (car *jqp-transient-activation-holding*)))) s-act)
		     (setf (nth 2 (assoc wme (car *jqp-transient-activation-holding*))) s-act))
		    ((and (null (nth 3 (assoc wme (car *jqp-transient-activation-holding*)))) sac-act)
		     (setf (nth 3 (assoc wme (car *jqp-transient-activation-holding*))) sac-act))
		    ((and (null (nth 4 (assoc wme (car *jqp-transient-activation-holding*)))) t-act)		  
		     (setf (nth 4 (assoc wme (car *jqp-transient-activation-holding*))) t-act))
		    ((and (nth 3 (assoc wme (car *jqp-transient-activation-holding*))) sac-act)			  
		     (setf (nth 3 (assoc wme (car *jqp-transient-activation-holding*))) sac-act))
		    (t *jqp-transient-activation-holding*)))
	  (progn (push (car *jqp-transient-activation-holding*) *jqp-act-out*)
		 (setf *jqp-transient-activation-holding* nil)
		 (push (list (list wme b-act s-act sac-act t-act pos neg) (list time)) 
		       *jqp-transient-activation-holding*)))
      (push (list (list wme b-act s-act sac-act t-act pos neg) (list time)) 
	    *jqp-transient-activation-holding*))))


(defun compute-base-level-activation (wme)
  (if (and (get-safe-slot-value wme 'positive) (get-safe-slot-value wme 'negative))
      (setf (wme-base-level wme)
	(+ *base-level-constant*
	   (compute-references (car (wme-references wme)) (cdr (wme-references wme))
			       (wme-creation-time wme) (- *base-level-learning*))))
    (setf (wme-base-level wme)
      (+ *base-level-constant*
	 (compute-references (car (wme-references wme)) (cdr (wme-references wme))
			     (wme-creation-time wme) (- *base-level-learning*)))))
  (jqp-get-transient-activation *time* :wme wme :b-act (wme-base-level wme))
  (signal-output *activation-trace* "   Computing a base level of ~6,3F from ~D references from creation time ~6,3F"
		 (wme-base-level wme) (round (first (wme-references wme))) (wme-creation-time wme))
  (wme-base-level wme))


(defmacro test-slots-pm (wme slots actions)
  "Tests that wme slots satisfy all the actions.  Equal test necessary to
   handle the full potential range of non-wme values.  Partial matching
   on wme values.  Returns the activation of wme if above threshold,
   nil otherwise."
  `(let ((activation (activation ,wme)))
     (jqp-get-transient-activation *time* :wme ,wme :s-act activation)
     (signal-output *partial-matching-trace* "Partial matching chunk ~S with activation ~6,3F"
		    ,wme activation)
     (dolist (action ,actions
	       (progn
		 (jqp-get-transient-activation *time* :wme ,wme :t-act activation)
		 (if (< (length *jqp-act-hold*) 3)
		     *jqp-act-hold*
		   (progn	
		     (jqp-get-transient-activation 
		      *time* :wme ,wme :sac-act (nth 1 *jqp-act-hold*))
		     (setf *jqp-act-hold* nil)))					  
		 (signal-output *partial-matching-trace* "Matching score of chunk ~S is ~6,3F."
				,wme activation)		       
		 activation))
       (let* ((value (slots-slot ,slots (action-slot action)))
	      (target (interpret-value action))
	      (similarity (get-similarity value target))
	      (negation (action-negation action)))
	 (signal-output *partial-matching-trace* "Similarity between chunks ~S and ~S is ~6,3F"
			value target similarity)
	 (cond ((null negation)
		(if (numberp similarity)
		    (let ((match (* similarity *mismatch-penalty*)))
		      (incf activation match)
		      (push activation *jqp-act-hold*)
		      (jqp-get-transient-activation 
		       *time* :wme ,wme :sac-act activation)
		      (signal-output *partial-matching-trace* "Adjusting activation by ~6,3F to ~6,3F"
				     match activation))
		  (unless similarity
		    (signal-output *partial-matching-trace* "Non-chunk mismatch between ~S and ~S: failure."
				   value target)
		    (return nil))))
	       ((eq negation '-)
		(if (numberp similarity)
		    (when (= similarity *max-sim*)
		      (let ((match (* *max-dif* *mismatch-penalty*)))
			(incf activation match)				
			(signal-output *partial-matching-trace* "Adjusting activation by ~6,3F to ~6,3F"
				       match activation)))
		   (when similarity
		     (signal-output *partial-matching-trace* "Non-chunk match between ~S and ~S for negation test: failure."
				    value target)
		     (return nil))))
	      ((fboundp negation)
	       (unless (and (numberp value) (numberp target)
			    (funcall negation value target))
		 (signal-output *exact-matching-trace* "Value ~S does not match test ~S with condition ~S: test fails."
				value negation target)
		 (return nil)))
	      (t
	       (signal-output *exact-matching-trace* "Unknown test ~S: ignoring it."
			      negation)))))))


;;
;; Evaluation Construction
;;

(defun jqp-get-activation-fct (wme &key (level 3))
  "Returns transient activation levels of wme. As default, it returns 
level 2 activation, baseline + spreading activation + affective priming effect."
  (if (listp wme)
      (let ((results nil))
	(dolist (i wme (reverse results))
	  (setf i (get-safe-wme i))
	  (when i
	    (let ((activation (nth level (assoc i (car *jqp-transient-activation-holding*)))))
	      (push activation results)))))
      (nth level (assoc (get-safe-wme wme) (car *jqp-transient-activation-holding*)))))


(defun jqp-sort (x y)
  "Given two lists x and y, it returns a (list (x1 y1) ...) 
such that x1 < x2 < ...."
  (let* ((len (length x)) (result nil)) 
    (dotimes (i len (sort result #'< :key #'car))
      (when (> (nth i x) *retrieval-threshold*)
	(push (list (nth i x) (nth i y)) result)))))


(defun jqp-convert-to-weight (x)
  "It normalizes raw activation levels with activation levels 
less than *retrieval-threshold* excluded."
  (cond ((or (numberp x) (stringp x))
	 (error (format t "non list or string can not be converted into weights.")))
	((listp x) 
	 (let* ((x (mapcar #'(lambda (n) 
		      (if (> n *retrieval-threshold*) n 0)) x))
		(sum (jqp-sum x)))
	   (mapcar #'(lambda (n) (if (<= sum 0) 0
				   (/ n sum))) x))))) 
 

(defun jqp-weighted-attitude-list (w x)
  "Given normalized activation levels w and valence x, it returns 
their average affect weighted by respective normalized activation level."
  (let* ((sresult (jqp-sort w x))
	 (num (length sresult)) 
	 (len (if (<= (- (length sresult) num) 0) 
		  0 (- (length sresult) num)))
	 (tresult (reverse (nthcdr len sresult)))) 
    (mapcar #'(lambda (n) (* (first n) (second n))) tresult))) 


(defun jqp-pos-attitude-colorization-fct (wmes &key slot-value)
  "It implements Eq (3) for positive affect. It returns the constructed 
positive affect of an object based on *delta* and the average positive 
affect of its associated chunks (considerations) weighted by their 
respective activation level. It takes a list as argument whose 1st element 
is the object of evaluation and the rest being other sources of spreading 
activation and affective priming effect, i.e., those chunks currently 
held in WM."
  (let* ((obj (list (first wmes)))
	 (obj-scaled-activation 1) ; always 1 and made explicit
	 (obj-paffect (car (jqp-get-paffect-associated-chunks-fct obj)))
	 (considerations-source wmes) 
	 (considerations (remove (get-safe-wme (car obj)) 
				 (jqp-get-considerations-on-the-spot considerations-source :slot-value slot-value)))
	 (considerations-activation (jqp-get-activation-fct considerations))
	 (considerations-scaled-activation 
	  (jqp-convert-to-weight considerations-activation)) 
	 (considerations-paffect (jqp-get-paffect-associated-chunks-fct considerations :slot-value slot-value)))
    (let* ((weighted-obj-attitude (* (- 1 *delta*) obj-scaled-activation obj-paffect))
	   (weighted-considerations-activation 
	    (mapcar #'(lambda (n) (* *delta* n))
		    considerations-scaled-activation))
	   (weighted-considerations-attitude 
	    (jqp-weighted-attitude-list weighted-considerations-activation considerations-paffect)))	     
      (+ weighted-obj-attitude (jqp-sum weighted-considerations-attitude)))))


(defun jqp-neg-attitude-colorization-fct (wmes &key slot-value)
  "It implements Eq (3) for negative affect. It returns the constructed 
negative affect of an object based on *delta* and the average negative 
affect of its associated chunks (considerations) weighted by their 
respective activation level. It takes a list as argument whose 1st element 
is the object of evaluation and the rest being other sources of spreading 
activation and affective priming effect, i.e., other chunks currently 
held in WM."
  (let* ((obj (list (first wmes)))
	 (obj-scaled-activation 1) ; always 1 and included to make it explicitw
	 (obj-naffect (car (jqp-get-naffect-associated-chunks-fct obj)))
	 (considerations-source wmes) ; object + other wmes currently in working memory
	 (considerations (remove (get-safe-wme (car obj)) 
				 (jqp-get-considerations-on-the-spot considerations-source :slot-value slot-value)))
	 (considerations-activation (jqp-get-activation-fct considerations))
	 (considerations-scaled-activation 
	  (jqp-convert-to-weight considerations-activation))
	 (considerations-naffect (jqp-get-naffect-associated-chunks-fct considerations :slot-value slot-value)))
    (let* ((weighted-obj-attitude (* (- 1 *delta*) obj-scaled-activation obj-naffect))
	   (weighted-considerations-activation 
	    (mapcar #'(lambda (n) (* *delta* n))
		    considerations-scaled-activation))
	   (weighted-considerations-attitude 
	    (jqp-weighted-attitude-list weighted-considerations-activation considerations-naffect)))	     
      (+ weighted-obj-attitude (jqp-sum weighted-considerations-attitude)))))


(defun jqp-attitude-colorization-fct (wmes &key slot-value)
  "It implements Eq (3). It returns the constructed overal affect towards 
an object based on *delta* and the average positive affect of its associated 
chunks (considerations) weighted by their respective activation. It takes 
a list as argument whose 1st element is the object of evaluation and the 
rest being other sources of spreading activation and affective priming effect, 
i.e., other chunks currently held in WM."
  (let ((x (jqp-pos-attitude-colorization-fct wmes :slot-value slot-value))
        (y (jqp-neg-attitude-colorization-fct wmes :slot-value slot-value)))
    (+ y x)))


(defmacro jqp-attitude-colorization (wmes)
  `(jqp-attitude-colorization-fct ',wmes))


(defun jqp-increase-wme-references (wme)
  (let* ((wme (if (wmep wme) wme (get-safe-wme wme))))	
    (incf (car (wme-references wme)))))

(defun jqp-increase-considerations-on-spot-references (wmes &key slot-value)
  "It increases the references of chunks used to construct the attitude of an object. 
It takes a list as argument whose 1st element is the object of evaluation and the 
rest being other sources of spreading activation and affective priming effect, 
i.e., other chunks currently held in WM."
  (let* ((obj (list (first wmes)))
	 (considerations-source wmes)	
	 (considerations 
	  (remove (get-safe-wme (car obj)) 
		  (jqp-get-considerations-on-the-spot considerations-source :slot-value slot-value)))
	 (considerations-activation (jqp-get-activation-fct considerations))
	 (considerations-activation-alist (jqp-sort considerations-activation considerations)) 
       	 (num (length considerations-activation-alist))
	 (acutal-considerations-activation-alist 
	  (let ((len (if (<= (- (length considerations-activation-alist) num) 0) 
			 0 (- (length considerations-activation-alist) num)))) 
	    (reverse (nthcdr len considerations-activation-alist))))) 
    (dolist (i acutal-considerations-activation-alist)
      (jqp-increase-wme-references (nth 1 i)))))




;;;
;;;
;;; Online Updating
;;;
;;;


(defun jqp-get-number-of-association (wmei wmej)
  "It returns the number of associations from wmei to wmej 
when wmei is the slot of goal in a procedural knowledge that 
is fired."
  (let* ((wmei (get-safe-wme wmei))
	 (wmej (get-safe-wme wmej)))
    (if (get-ia wmei wmej) (ia-count (get-ia wmei wmej)) 0)))


(defun jqp-get-ol-order (wme)
  "It retrieves the total number of associations from init-ol-order to wme. 
init-ol-order is a system chunck whose association to wme increases whenever 
the ol tally of a wme is updated."
  (jqp-get-number-of-association 'init-ol-order wme))


(defun online-updating (obj x y)
  "It represents Eq (4), on-line updating mechanism, for postive 
and negative affect separately."
  (let* ((wmei (get-safe-wme obj))	
	 (olorder (jqp-get-ol-order wmei))) 
    (+ x (* (expt *rho* olorder) y))))


;;;
;;;
;;; Report Explicit Evaluation and Activation (Perception)
;;;
;;;


(defun jqp-attitude-report (wme)
  (let* ((obj (if (listp wme) (first wme) wme))
	 (attitude (jqp-attitude-colorization-fct (list obj))))
    (format t "~&I think ~S is ~S!!" obj attitude)))


(defun jqp-record-survey-response-evaluation (wme)
  (let ((pos (jqp-pos-attitude-colorization-fct wme))
	(neg (jqp-neg-attitude-colorization-fct wme)))
    (push (list *time* (car wme) (+ pos neg)) *survey-response-evaluation*)))


(defun jqp-perception-report (wme)
  (let* ((trait (if (listp wme) (first wme) wme))
	 (obj (if (listp wme) (second wme)))
	 (s-act (jqp-get-activation-fct trait)))
    (format t "~&I think ~S is related to ~S by ~d!!" obj trait s-act)))

(defun jqp-record-survey-response-perception (wmes)
  (let* ((wmei (first wmes))	
	 (wmej (second wmes))	
	 (bact (jqp-get-activation-fct wmei :level 1))
	 (bact1 (jqp-get-activation-fct wmei :level 2))
	 (sact (jqp-get-activation-fct wmei :level 3)))
    (push (list *time* wmej wmei bact bact1 sact)
	  *survey-response-perception*)))



;;;
;;; Reset
;;;

(defun jqp-reset ()
  "It resets jqp into its initial state except parameter values. 
It is used for simulation when parameter values are varied."
  (setf *jqp-act-hold* nil)
  (setf *jqp-act-out* nil)
  (setf *jqp-out* nil)
  (setf *jqp-transient-activation-holding* nil)
  (setf *survey-response-evaluation* nil)
  (setf *survey-response-perception* nil)
  (reset))
